/* CS 472/672 Assignment #2 - main.c file */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "rcm.h" 
#include "mindeg.h" 
#include "matrhs.h"
#include "dirsolver.h"


int main(int argc, char **args)
{ 
/*
   Variable definitions:

      ia,ja,a    usual ia-ja arrays
      b          right hand side (noisy image)
      x          solution (original ordering)

      order      ordering vector: order[new_order] = old_order
      invord     inverse of order:  invord[order[i]] = i
      norder     =0 Natural ordering
                 =1 RCM ordering
                 =2 Min degree ordering

      LU         LU matrix in sparse row format

      n          the grid size is nxn
      N          number of unknowns
      nnz        number of nonzeros in LU matrix

      alpha      parameter for image denoising
      iter       number of denoising iteration

      imagefile  output file for the smoothed image

*/

   char    *usageMessage =
           "Usage: ./programOpt n ordering   or   ./programDebug n ordering\n"
           "       n = grid size in each dimension for the image.\n"
           "       ordering = 0 for original ordering, 1 for RCM, 2 for min deg.\n";
   int     n, N, norder, nnz, i, k, iter;
   int     *ia, *ja, *order, *invord; 
   double  *a, *x, *b, alpha;
   FILE    *imagefile;
   clock_t afterProblem, afterOrder, afterAnalyze, afterFactor, afterSolve;
   struct sparse_row *LU;


   /* Set up parameters. */
   n=4; norder=0; 

   if (argc==1) {printf("%s",usageMessage); exit(0);}
   if (argc>1) sscanf(args[1],"%d",&n);
   if (argc>2) sscanf(args[2],"%d",&norder);

   N = n*n;
   alpha = 3e-5;
   iter = 40;
   printf("Solving the Image Denoising Problem\n");
   printf("Grid size = %d x %d\n",n,n);
   printf("Ordering method = %s\n\n",((norder==0) ? "Natural" : (norder==1) ? "RCM" : (norder==2) ? "Minimum degree" : "Error"));


   /* Allocate space for arrays ia, ja, a, x, and b.             */
   /* Allocate space for order and inverse order vectors.        */
   /* Allocate array of sparse_row structures for the LU matrix, */
   /* each structue points to a row.                             */
   ia = (int *) malloc((N+1)*sizeof(int));
   ja = (int *) malloc(5*N*sizeof(int));
   a = (double *) malloc(5*N*sizeof(double));
   x = (double *) malloc(N*sizeof(double));
   b = (double *) malloc(N*sizeof(double));
   order = (int *) malloc(N*sizeof(int));
   invord = (int *) malloc(N*sizeof(int));
   LU = (struct sparse_row *) malloc(N*sizeof(struct sparse_row));


   /* Set up image matrix and RHS (noisy image) */
   imgmatrix(n, alpha, ia, ja, a);
   noisyimg(n, b);

   afterProblem = clock();


   /*  Ordering and inverse ordering. */
   if (norder==0) for (i=0; i<N; i++) order[i] = i;
   if (norder==1) rcm(ia, ja, N, order);
   if (norder==2) mindeg(ia, ja, N, order);
   if (norder>2) {printf("Error in norder\n"); exit(0);}

   for (i=0; i<N; i++) invord[order[i]] = i;

   afterOrder = clock();
   printf("Time after ordering = %12.4f\n",
          ((double) afterOrder / (double) CLOCKS_PER_SEC));


   /* Symbolic factorization (analyze phase) */
   sfac_smart(ia, ja, N, order, invord, &nnz, LU);

   afterAnalyze = clock();
   printf ("Time after analyze phase = %12.4f\n"
           "Number of non-zeroes in LU = %d\n",
           ((double) afterAnalyze / (double) CLOCKS_PER_SEC), nnz);


   /* Numeric factorization (factor phase) */
   factor(N, ia, ja, a, order, invord, LU);

   afterFactor = clock();
   printf("Time after factor phase = %12.4f\n",
          ((double) afterFactor / (double) CLOCKS_PER_SEC));


   /* Image denoising */
   for (k=0; k<iter; k++) {
      solve(N, x, b, order, LU);
      for (i=0; i<N; i++) b[i] = x[i];
   }

   afterSolve = clock();
   printf("Time after image denoising: %12.4f\n\n",
          ((double) afterSolve / (double) CLOCKS_PER_SEC));


   /* Save the denoised image to imagefile */
   imagefile=fopen("img.m","w");
   for (i=0; i<N; i++) fprintf(imagefile,"%25.15e\n",x[i]);
   fclose(imagefile);


   /* Report the total solution times */
   printf("Time for each phase:\n"
          "            Problem setup = %12.4f\n"
          "              Re-ordering = %12.4f\n"
          " Analysis/Symbolic Factor = %12.4f\n"
          "           Numeric Factor = %12.4f\n"
          "  (40) Forward/Back Solve = %12.4f\n\n",
        ( (double) afterProblem                   / (double) CLOCKS_PER_SEC ),
        ( (double) (afterOrder   - afterProblem ) / (double) CLOCKS_PER_SEC ),
        ( (double) (afterAnalyze - afterOrder   ) / (double) CLOCKS_PER_SEC ),
        ( (double) (afterFactor  - afterAnalyze ) / (double) CLOCKS_PER_SEC ),
        ( (double) (afterSolve   - afterFactor  ) / (double) CLOCKS_PER_SEC ) );


   /* Free memory  */
   clean(N, LU);   
   free(ia); free(ja); free(a); free(x); free(b);
   free(order); free(invord);

   return 0;
}
